"""Benchmarks baselines used by the OpenSTEF Beam benchmarking utilities.

This package exposes baseline forecasters for use in backtesting.
"""

# SPDX-FileCopyrightText: 2025 Contributors to the OpenSTEF project <openstef@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from openstef_beam.benchmarking.baselines.openstef4 import (
    OpenSTEF4BacktestForecaster,
    WorkflowCreationContext,
    create_openstef4_preset_backtest_forecaster,
)

__all__ = [
    "OpenSTEF4BacktestForecaster",
    "WorkflowCreationContext",
    "create_openstef4_preset_backtest_forecaster",
]
