# SPDX-FileCopyrightText: 2025 Contributors to the OpenSTEF project <openstef@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Data models and types for benchmark targets and configurations."""

from openstef_beam.benchmarking.models.benchmark_target import BenchmarkTarget

__all__ = ["BenchmarkTarget"]
